# This file is part of SSHare.
#
# SSHare is free software: you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# SSHare is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# SSHare. If not, see <https://www.gnu.org/licenses/>.

import importlib
import importlib.util

from sshare.plugins.config import Flag
from sshare.plugins.config import NoDefault

class PluginManager:
    def __init__(self, types, logger, config, arg_parser):
        self._uninitialized = []
        self._logger        = logger
        self._config        = config
        self._arg_parser    = arg_parser

        class PluginState:
            def __init__(self):
                self.active = []
                self.inactive = []
        for type in types:
            setattr(self, type, PluginState())

    def add_from(self, location, *args, **kwargs):
        for plugin in args:
            plugin = location(plugin, self._config)
            plugin.set_logger(self._logger)
            plugin.add_args(self._arg_parser)
            self._uninitialized.append(plugin)

    def activate(self, activate_type=None):
        args = self._arg_parser.parse_args()
        for plugin in self._uninitialized.copy():
            if activate_type == None or activate_type in plugin.plugin_type:
                self._uninitialized.remove(plugin)
                active = plugin.load_args_and_activate(args)
                for type in plugin.plugin_type:
                    getattr(
                        getattr(self, type),
                        active[type],
                    ).append(plugin)

class Plugin:
    def __init__(self, name, module, external_config):
        self.__dict__ = module.__dict__
        self.name     = name

        if not isinstance(self.plugin_type, set):
            self.plugin_type = { self.plugin_type }

        if hasattr(self, "activate"):
            if not isinstance(self.activate, dict):
                activate = self.activate
                self.activate = dict()
                for plugin_type in self.plugin_type:
                    self.activate[plugin_type] = activate
        else:
            self.activate = dict()
            for plugin_type in self.plugin_type:
                self.activate[plugin_type] = set()

        if hasattr(self, "config"):
            config = self.config
        else:
            config = dict()
        if external_config == None:
            external_config = dict()
        for key in config.keys():
            if key in external_config:
                config[key] = external_config[key]

        if hasattr(self, "args"):
            for arg in self.args.items():
                arg[1].bind(self, arg[0])
                value = arg[1].default()
                if value != NoDefault:
                    config[arg[0]] = value
        else:
            self.args = dict()

        class Config: pass
        flat_config = Config()
        flat_config.__dict__ = config
        self.config = flat_config

    def set_logger(self, logger):
        self.logger = logger

    def add_args(self, arg_parser):
        for arg in self.args.values():
            arg.add(arg_parser)

    def load_args_and_activate(self, args):
        passed_args = set()
        for arg_name, arg in self.args.items():
            was_set, value = arg.extract(args)
            if was_set:
                if value != Flag:
                    setattr(self.config, arg_name, value)
                passed_args.add(arg_name)
        activate = dict()
        run_init = False
        for type in self.plugin_type:
            if self.activate[type] <= passed_args:
                activate[type] = "active"
                run_init = True
            else:
                activate[type] = "inactive"
        if run_init and hasattr(self, "init"):
            self.init()
        return activate
    
    @staticmethod
    def internal(location, name=None, config=None):
        def _load_internal(_name, _config):
            return Plugin(_name, importlib.import_module(f"{location}.{_name}"), _config.get(_name, dict()))
        if name == None:
            return _load_internal
        else:
            return _load_internal(name, config)

    @staticmethod
    def external(path, config):
        module_spec = importlib.util.spec_from_file_location(
            path.stem,
            path.as_posix(),
        )
        module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module)
        return Plugin(path.stem, module, config.get(path.stem, dict()))
